#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import unicode_literals

from io import StringIO
import json

from py2neo.core import Node, Rel, Rev, Path, Relationship
from py2neo.lang import Writer
from py2neo.util import is_collection, ustr, xstr


__all__ = list(map(xstr, ["CypherParameter", "CypherWriter", "cypher_escape", "cypher_repr"]))


class CypherParameter(object):
    """ Object representing a parameter within a Cypher statement.
    """

    def __init__(self, key, value=None):
        self.key = key
        self.value = value


class CypherWriter(Writer):
    """ Writer for Cypher data. This can be used to write to any
    file-like object, such as standard output::

        >>> from py2neo import Node
        >>> from py2neo.cypher import CypherWriter
        >>> import sys
        >>> writer = CypherWriter(sys.stdout)
        >>> writer.write(Node("Person", name="Alice"))
        (:Person {name:"Alice"})

    """

    safe_first_chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_"
    safe_chars = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_"

    default_sequence_separator = ","
    default_key_value_separator = ":"

    def __init__(self, file=None, **kwargs):
        Writer.__init__(self, file)
        self.sequence_separator = kwargs.get("sequence_separator", self.default_sequence_separator)
        self.key_value_separator = \
            kwargs.get("key_value_separator", self.default_key_value_separator)

    def write(self, obj):
        """ Write any entity, value or collection.
        """
        if obj is None:
            pass
        elif isinstance(obj, Node):
            self.write_node(obj)
        elif isinstance(obj, Rel):
            self.write_rel(obj)
        elif isinstance(obj, Relationship):
            self.write_relationship(obj)
        elif isinstance(obj, Path):
            self.write_path(obj)
        elif isinstance(obj, CypherParameter):
            self.write_parameter(obj)
        elif isinstance(obj, dict):
            self.write_map(obj)
        elif is_collection(obj):
            self.write_list(obj)
        else:
            self.write_value(obj)

    def write_value(self, value):
        """ Write a value.
        """
        self.file.write(ustr(json.dumps(value, ensure_ascii=False)))

    def write_identifier(self, identifier):
        """ Write an identifier.
        """
        if not identifier:
            raise ValueError("Invalid identifier")
        identifier = ustr(identifier)
        safe = (identifier[0] in self.safe_first_chars and
                all(ch in self.safe_chars for ch in identifier[1:]))
        if not safe:
            self.file.write("`")
            self.file.write(identifier.replace("`", "``"))
            self.file.write("`")
        else:
            self.file.write(identifier)

    def write_label(self, label):
        """ Write a label.
        """
        self.write_literal(":")
        self.write_identifier(label)

    def write_list(self, collection):
        """ Write a list.
        """
        self.file.write("[")
        link = ""
        for value in collection:
            self.file.write(link)
            self.write(value)
            link = self.sequence_separator
        self.file.write("]")

    def write_literal(self, text):
        """ Write literal text.
        """
        self.file.write(ustr(text))

    def write_map(self, mapping):
        """ Write a map.
        """
        self.file.write("{")
        link = ""
        for key, value in sorted(mapping.items()):
            self.file.write(link)
            self.write_identifier(key)
            self.file.write(self.key_value_separator)
            self.write(value)
            link = self.sequence_separator
        self.file.write("}")

    def write_node(self, node, name=None, properties=None):
        """ Write a node.
        """
        self.file.write("(")
        if name:
            self.write_identifier(name)
        if node is not None:
            for label in sorted(node.labels):
                self.write_label(label)
            if properties is None:
                if node.properties:
                    if name or node.labels:
                        self.file.write(" ")
                    self.write_map(node.properties)
            else:
                self.file.write(" ")
                self.write(properties)
        self.file.write(")")

    def write_rel(self, rel, name=None):
        """ Write a relationship (excluding nodes).
        """
        if isinstance(rel, Rev):
            self.file.write("<-[")
        else:
            self.file.write("-[")
        if name:
            self.write_identifier(name)
        self.file.write(":")
        self.write_identifier(rel.type)
        if rel.properties:
            self.file.write(" ")
            self.write_map(rel.properties)
        if isinstance(rel, Rev):
            self.file.write("]-")
        else:
            self.file.write("]->")

    def write_relationship(self, relationship, name=None):
        """ Write a relationship (including nodes).
        """
        self.write_node(relationship.start_node)
        self.write_rel(relationship.rel, name)
        self.write_node(relationship.end_node)

    def write_parameter(self, parameter):
        """ Write a parameter key in curly brackets.
        """
        self.file.write("{")
        self.file.write(ustr(parameter.key))
        self.file.write("}")

    def write_path(self, path):
        """ Write a :class:`py2neo.Path`.
        """
        nodes = path.nodes
        self.write_node(nodes[0])
        for i, rel in enumerate(path.rels):
            self.write_rel(rel)
            self.write_node(nodes[i + 1])


def cypher_escape(identifier):
    """ Escape a Cypher identifier in backticks.

    ::

        >>> cypher_escape("this is a `label`")
        '`this is a ``label```'

    """
    string = StringIO()
    writer = CypherWriter(string)
    writer.write_identifier(identifier)
    return string.getvalue()


def cypher_repr(obj):
    """ Generate the Cypher representation of an object.
    """
    string = StringIO()
    writer = CypherWriter(string)
    writer.write(obj)
    return string.getvalue()



from __future__ import print_function

import sys
sys.path.insert(1, '../py2neo/')

import argparse
import os

from py2neo.core import Graph
from py2neo.ext.spatial import Spatial


NEO_URL = "http://localhost:{port}/db/data/"
DEFAULT_DB = NEO_URL.format(port=7474)
DATA_HOME = 'examples/data'


parser = argparse.ArgumentParser(
    description='Load the example data into map layers')

parser.add_argument(
    'data', type=str, help='the name of the data file to load')


parser.add_argument(
    '--layer', dest='layer_name', action='store',
    help="""The layer to add the data to.
    This will be created if it does not already exist.""")

parser.add_argument(
    '--port', dest='port_address', action='store',
    help='The port Neo is running on')


def load(server_url, geometry_name, wkt_string, layer_name):
    graph = Graph(server_url)
    spatial = Spatial(graph)
    spatial.create_layer(layer_name)
    spatial.create_geometry(geometry_name, wkt_string, layer_name)
    print('done')


if __name__ == '__main__':
    args = parser.parse_args()

    data_file = args.data
    if data_file.endswith('wkt'):
        print('Just provide the name of the file - extension not required')

    geometry_name = data_file
    data_file += '.wkt'
    layer_name = args.layer_name
    neo_port = args.port_address

    server_url = NEO_URL.format(port=neo_port) if neo_port else DEFAULT_DB

    working_directory = os.path.dirname(os.path.realpath(__file__))
    ext_root = working_directory.strip('scripts')
    data_uri = os.path.join(ext_root, DATA_HOME, data_file)

    with open(data_uri, 'r') as fh:
        wkt_string = fh.read()

    load(
        server_url=server_url, geometry_name=geometry_name,
        wkt_string=wkt_string, layer_name=layer_name
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from py2neo.cypher.error.core import ClientError, DatabaseError, TransientError


class ArithmeticError(ClientError):
    """ Invalid use of arithmetic, such as dividing by zero.
    """


class ConstraintViolation(ClientError):
    """ A constraint imposed by the statement is violated by the data
    in the database.
    """


class EntityNotFound(ClientError):
    """ The statement is directly referring to an entity that does not
    exist.
    """


class InvalidArguments(ClientError):
    """ The statement is attempting to perform operations using invalid
    arguments
    """


class InvalidSemantics(ClientError):
    """ The statement is syntactically valid, but expresses something
    that the database cannot do.
    """


class InvalidSyntax(ClientError):
    """ The statement contains invalid or unsupported syntax.
    """


class InvalidType(ClientError):
    """ The statement is attempting to perform operations on values
    with types that are not supported by the operation.
    """


class NoSuchLabel(ClientError):
    """ The statement is referring to a label that does not exist.
    """


class NoSuchProperty(ClientError):
    """ The statement is referring to a property that does not exist.
    """


class ParameterMissing(ClientError):
    """ The statement is referring to a parameter that was not provided
    in the request.
    """


class ExecutionFailure(DatabaseError):
    """ The database was unable to execute the statement.
    """


class ExternalResourceFailure(TransientError):
    """ The external resource is not available
    """



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2013-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# HTTP/HTTPS Port Numbers

HTTP_PORT = 80
HTTPS_PORT = 443


# HTTP Status Codes <http://www.iana.org/assignments/http-status-codes>

# 1xx: Informational
# Request received, continuing process
CONTINUE = 100
SWITCHING_PROTOCOLS = 101
PROCESSING = 102

# 2xx: Success
# The action was successfully received, understood, and accepted
OK = 200
CREATED = 201
ACCEPTED = 202
NON_AUTHORITATIVE_INFORMATION = 203
NO_CONTENT = 204
RESET_CONTENT = 205
PARTIAL_CONTENT = 206
MULTI_STATUS = 207
ALREADY_REPORTED = 208
IM_USED = 226

# 3xx: Redirection
# Further action must be taken in order to complete the request
MULTIPLE_CHOICES = 300
MOVED_PERMANENTLY = 301
FOUND = 302
SEE_OTHER = 303
NOT_MODIFIED = 304
USE_PROXY = 305
TEMPORARY_REDIRECT = 307
PERMANENT_REDIRECT = 308

# 4xx: Client Error
# The request contains bad syntax or cannot be fulfilled
BAD_REQUEST = 400
UNAUTHORIZED = 401
PAYMENT_REQUIRED = 402
FORBIDDEN = 403
NOT_FOUND = 404
METHOD_NOT_ALLOWED = 405
NOT_ACCEPTABLE = 406
PROXY_AUTHENTICATION_REQUIRED = 407
REQUEST_TIMEOUT = 408
CONFLICT = 409
GONE = 410
LENGTH_REQUIRED = 411
PRECONDITION_FAILED = 412
REQUEST_ENTITY_TOO_LARGE = 413
REQUEST_URI_TOO_LONG = 414
UNSUPPORTED_MEDIA_TYPE = 415
REQUESTED_RANGE_NOT_SATISFIABLE = 416
EXPECTATION_FAILED = 417
UNPROCESSABLE_ENTITY = 422
LOCKED = 423
FAILED_DEPENDENCY = 424
UPGRADE_REQUIRED = 426
PRECONDITION_REQUIRED = 428
TOO_MANY_REQUESTS = 429
REQUEST_HEADER_FIELDS_TOO_LARGE = 431

# 5xx: Server Error
# The server failed to fulfill an apparently valid request
INTERNAL_SERVER_ERROR = 500
NOT_IMPLEMENTED = 501
BAD_GATEWAY = 502
SERVICE_UNAVAILABLE = 503
GATEWAY_TIMEOUT = 504
HTTP_VERSION_NOT_SUPPORTED = 505
VARIANT_ALSO_NEGOTIATES = 506
INSUFFICIENT_STORAGE = 507
NOT_EXTENDED = 510
NETWORK_AUTHENTICATION_REQUIRED = 511



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from py2neo.batch.core import *
from py2neo.batch.pull import *
from py2neo.batch.push import *
from py2neo.batch.read import *
from py2neo.batch.write import *


__all__ = [

    "BatchResource",

    "Batch",
    "WriteBatch",
    "PullBatch",
    "PushBatch",
    "ReadBatch",

    "Job",
    "JobResult",
    "Target",

    "AddNodeLabelsJob",
    "CreateNodeJob",
    "CreatePathJob",
    "CreateRelationshipJob",
    "CreateUniquePathJob",
    "CypherJob",
    "DeleteEntityJob",
    "DeletePropertiesJob",
    "DeletePropertyJob",
    "PullNodeLabelsJob",
    "PullPropertiesJob",
    "PullRelationshipJob",
    "PushNodeLabelsJob",
    "PushPropertiesJob",
    "PushPropertyJob",
    "RemoveNodeLabelJob",

    "BatchError",

]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import os
import sys

from py2neo.core import Relationship, Node, Path
from py2neo.cypher.error import CypherError, CypherTransactionError
from py2neo.env import NEO4J_URI
from py2neo.util import compact


HELP = """\
Usage: {script} [«options»] «statement» [ [«options»] «statement» ... ]

Execute a Cypher statement against a Neo4j database.

Formatting Options:
  -c --csv    write output as comma separated values
  -h --human  write output as human-readable text
  -j --json   write output as JSON (See http://json.org)
  -t --tsv    write output as tab separated values

Parameter Options:
  -f «parameter-file»
  -p «name» «value»

Environment:
  NEO4J_URI - base URI of Neo4j database, e.g. http://localhost:7474

Report bugs to nigel@py2neo.org
"""


def _help(script):
    sys.stderr.write(HELP.format(script=os.path.basename(script)))
    sys.stderr.write("\n")


class CypherCommandLine(object):

    def __init__(self, graph):
        self.parameters = {}
        self.parameter_filename = None
        self.graph = graph
        self.tx = None

    def begin(self):
        self.tx = self.graph.cypher.begin()

    def set_parameter(self, key, value):
        try:
            self.parameters[key] = json.loads(value)
        except ValueError:
            self.parameters[key] = value

    def set_parameter_filename(self, filename):
        self.parameter_filename = filename

    def execute(self, statement):
        import codecs
        if self.parameter_filename:
            columns = None
            with codecs.open(self.parameter_filename, encoding="utf-8") as f:
                for line in f:
                    line = line.strip()
                    if columns is None:
                        columns = line.split(",")
                    elif line:
                        values = json.loads("[" + line + "]")
                        p = dict(self.parameters)
                        p.update(zip(columns, values))
                        self.tx.append(statement, p)
        else:
            self.tx.append(statement, self.parameters)
        return self.tx.process()

    def commit(self):
        self.tx.commit()


def dehydrate(value):
    if isinstance(value, list):
        out = list(map(dehydrate, value))
    elif isinstance(value, Node):
        out = {
            "self": value.ref,
            "properties": value.properties,
            "labels": list(value.labels),
        }
    elif isinstance(value, Relationship):
        out = {
            "self": value.ref,
            "properties": value.properties,
            "start": value.start_node.ref,
            "type": value.type,
            "end": value.end_node.ref,
        }
    elif isinstance(value, Path):
        out = [dehydrate(relationship) for relationship in value.relationships]
    else:
        out = value
    return out


def main():
    import sys
    from py2neo.core import ServiceRoot
    script, args = sys.argv[0], sys.argv[1:]
    if not args:
        _help(script)
        sys.exit(0)
    uri = NEO4J_URI.resolve("/")
    service_root = ServiceRoot(uri.string)
    out = sys.stdout
    output_format = None
    command_line = CypherCommandLine(service_root.graph)
    command_line.begin()
    while args:
        arg = args.pop(0)
        if arg.startswith("-"):
            if arg in ("-p", "--parameter"):
                key = args.pop(0)
                value = args.pop(0)
                command_line.set_parameter(key, value)
            elif arg in ("-f",):
                command_line.set_parameter_filename(args.pop(0))
            elif arg in ("-h", "--human"):
                output_format = None
            elif arg in ("-j", "--json"):
                output_format = "json"
            elif arg in ("-c", "--csv"):
                output_format = "csv"
            elif arg in ("-t", "--tsv"):
                output_format = "tsv"
            else:
                raise ValueError("Unrecognised option %s" % arg)
        else:
            try:
                results = command_line.execute(arg)
            except CypherError as error:
                sys.stderr.write("%s: %s\n\n" % (error.__class__.__name__, error.args[0]))
            else:
                for result in results:
                    if output_format == "json":
                        records = [compact(dict(zip(result.columns, map(dehydrate, record))))
                                   for record in result]
                        out.write(json.dumps(records, ensure_ascii=False, sort_keys=True, indent=2))
                        out.write("\n")
                    elif output_format in ("csv", "tsv"):
                        separator = "," if output_format == "csv" else "\t"
                        out.write(separator.join(json.dumps(column, separators=",:",
                                                            ensure_ascii=False)
                                                 for column in result.columns))
                        out.write("\n")
                        for record in result:
                            out.write(separator.join(json.dumps(dehydrate(value),
                                                                separators=",:",
                                                                ensure_ascii=False,
                                                                sort_keys=True)
                                                     for value in record))
                            out.write("\n")
                    else:
                        out.write(repr(result))
                    out.write("\n")
    try:
        command_line.commit()
    except CypherTransactionError as error:
        sys.stderr.write(error.args[0])
        sys.stderr.write("\n")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os

from py2neo.packages.httpstream.packages.urimagic import URI


__all__ = ["NEO4J_AUTH_TOKEN", "NEO4J_DIST", "NEO4J_HOME", "NEO4J_URI"]


#: Auth token for use in Neo4j 2.2 and above.
NEO4J_AUTH_TOKEN = os.getenv("NEO4J_AUTH_TOKEN", None)

#: Base URI for downloading Neo4j distribution archives.
NEO4J_DIST = os.getenv("NEO4J_DIST", "http://dist.neo4j.org/")

#: Default path for GraphServer instances.
NEO4J_HOME = os.getenv("NEO4J_HOME", ".")

#: Default URI for ServiceRoot instances.
NEO4J_URI = URI(os.getenv("NEO4J_URI", "http://localhost:7474/"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import logging

from py2neo.core import NodePointer, Service
from py2neo.cypher import RecordList
from py2neo.error import GraphError, Finished
from py2neo.packages.jsonstream import assembled, grouped
from py2neo.packages.httpstream.packages.urimagic import percent_encode, URI
from py2neo.util import pendulate, ustr, raise_from


__all__ = ["BatchError", "BatchResource", "Target", "Job", "JobResult", "CypherJob", "Batch"]


log = logging.getLogger("py2neo.batch")


class BatchError(GraphError):
    """ Wraps a base :class:`py2neo.GraphError` within a batch context.
    """

    batch = None
    job_id = None
    status_code = None
    uri = None
    location = None

    def __init__(self, message, batch, job_id, status_code, uri, location=None, **kwargs):
        GraphError.__init__(self, message, **kwargs)
        self.batch = batch
        self.job_id = job_id
        self.status_code = status_code
        self.uri = uri
        self.location = location


class BatchResource(Service):
    """ Resource for batch execution.
    """

    __instances = {}

    def __new__(cls, uri):
        try:
            inst = cls.__instances[uri]
        except KeyError:
            inst = super(BatchResource, cls).__new__(cls)
            inst.bind(uri)
            cls.__instances[uri] = inst
        return inst

    def post(self, batch):
        """ Post a batch of jobs to the server and receive a raw
        response.

        :arg batch: A :class:`.Batch` of jobs.
        :rtype: :class:`httpstream.Response`

        """
        num_jobs = len(batch)
        plural = "" if num_jobs == 1 else "s"
        log.info("> Sending batch request with %s job%s", num_jobs, plural)
        data = []
        for i, job in enumerate(batch):
            if job.finished:
                raise Finished(job)
            else:
                job.finished = True
            log.info("> {%s} %s", i, job)
            data.append(dict(job, id=i))
        response = self.resource.post(data)
        log.info("< Received batch response for %s job%s", num_jobs, plural)
        return response

    def run(self, batch):
        """ Execute a collection of jobs and discard any response.

        :arg batch: A :class:`.Batch` of jobs.

        """
        response = self.post(batch)
        log.info("< Discarding batch response")
        response.close()

    def stream(self, batch):
        """ Execute a collection of jobs and yield results as received.

        :arg batch: A :class:`.Batch` of jobs.
        :rtype: generator

        """
        response = self.post(batch)
        try:
            for i, result_data in grouped(response):
                result = JobResult.hydrate(assembled(result_data), batch)
                log.info("< %s", result)
                yield result
        finally:
            response.close()

    def submit(self, batch):
        """ Execute a collection of jobs and return all results.

        :arg batch: A :class:`.Batch` of jobs.
        :rtype: :class:`list`

        """
        response = self.post(batch)
        try:
            results = []
            for result_data in response.content:
                result = JobResult.hydrate(result_data, batch)
                log.info("< %s", result)
                results.append(result)
            return results
        except ValueError:
            # Here, we're looking to gracefully handle a Neo4j server bug
            # whereby a response is received with no content and
            # 'Content-Type: application/json'. Given that correct JSON
            # technically needs to contain {} at minimum, the JSON
            # parser fails with a ValueError.
            if response.content_length == 0:
                from sys import exc_info
                from traceback import extract_tb
                type_, value, traceback = exc_info()
                for filename, line_number, function_name, text in extract_tb(traceback):
                    if "json" in filename and "decode" in function_name:
                        return []
            raise
        finally:
            response.close()


class Target(object):
    """ A callable target for a batch job. This may refer directly to a URI
    or to an object that can be resolved to a URI, such as a :class:`py2neo.Node`.
    """

    #: The entity behind this target.
    entity = None

    #: Additional path segments to append to the resolved URI.
    segments = None

    def __init__(self, entity, *segments):
        self.entity = entity
        self.segments = segments

    @property
    def uri_string(self):
        """ The fully resolved URI string for this target.

        :rtype: string

        """
        if isinstance(self.entity, int):
            uri_string = "{{{0}}}".format(self.entity)
        elif isinstance(self.entity, NodePointer):
            uri_string = "{{{0}}}".format(self.entity.address)
        else:
            try:
                uri_string = self.entity.ref
            except AttributeError:
                uri_string = ustr(self.entity)
        if self.segments:
            if not uri_string.endswith("/"):
                uri_string += "/"
            uri_string += "/".join(map(percent_encode, self.segments))
        return uri_string


class Job(object):
    """ A single request for inclusion within a :class:`.Batch`.
    """

    #: The graph for which this job is intended (optional).
    graph = None

    #: Indicates whether or not the result should be
    #: interpreted as raw data.
    raw_result = False

    #: The HTTP method for the request.
    method = None

    #: A :class:`.Target` object used to determine the destination URI.
    target = None

    #: The request payload.
    body = None

    #: Indicates whether the job has been submitted.
    finished = False

    def __init__(self, method, target, body=None):
        self.method = method
        self.target = target
        self.body = body

    def __repr__(self):
        parts = [self.method, self.target.uri_string]
        if self.body is not None:
            parts.append(json.dumps(self.body, separators=",:"))
        return " ".join(parts)

    def __eq__(self, other):
        return self is other

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(id(self))

    def __iter__(self):
        yield "method", self.method
        yield "to", self.target.uri_string
        if self.body is not None:
            yield "body", self.body


class JobResult(object):
    """ The result returned from the server for a single
    :class:`.Job` following a :class:`.Batch` submission.
    """

    @classmethod
    def hydrate(cls, data, batch):
        graph = getattr(batch, "graph", None)
        job_id = data["id"]
        uri = data["from"]
        status_code = data.get("status")
        location = data.get("location")
        if graph is None or batch[job_id].raw_result:
            body = data.get("body")
        else:
            body = None
            try:
                body = graph.hydrate(data.get("body"))
            except GraphError as error:
                message = "Batch job %s failed with %s\n%s" % (
                    job_id, error.__class__.__name__, ustr(error))
                raise_from(BatchError(message, batch, job_id, status_code, uri, location), error)
            else:
                # If Cypher results, reduce to single row or single value if possible
                if isinstance(body, RecordList):
                    num_rows = len(body)
                    if num_rows == 0:
                        body = None
                    elif num_rows == 1:
                        body = body[0]
                        num_columns = len(body)
                        if num_columns == 1:
                            body = body[0]
        return cls(batch, job_id, uri, status_code, location, body)

    #: The :class:`.Batch` from which this result was returned.
    batch = None

    #: The unique ID for this job within the batch.
    job_id = None

    #: The URI destination of the original job.
    uri = None

    #: The status code returned for this job.
    status_code = None

    #: The ``Location`` header returned for this job (if included).
    location = None

    #: The response content for this job.
    content = None

    def __init__(self, batch, job_id, uri, status_code=None, location=None, content=None):
        self.batch = batch
        self.job_id = job_id
        self.uri = URI(uri)
        self.status_code = status_code or 200
        self.location = URI(location)
        self.content = content

    def __repr__(self):
        parts = ["{" + ustr(self.job_id) + "}", ustr(self.status_code)]
        if self.content is not None:
            parts.append(repr(self.content))
        return " ".join(parts)

    @property
    def graph(self):
        """ The corresponding graph for this result.

        :rtype: :class:`py2neo.Graph`

        """
        return self.batch.graph

    @property
    def job(self):
        """ The original job behind this result.

        :rtype: :class:`.Job`

        """
        return self.batch[self.job_id]


class CypherJob(Job):
    """ A Cypher execution job for inclusion within a batch. Consider using
    `Cypher transactions <py2neo.cypher.CypherTransaction>`_ instead of
    batched Cypher jobs.
    """

    target = Target("cypher")

    def __init__(self, statement, parameters=None):
        body = {"query": ustr(statement)}
        if parameters:
            body["params"] = dict(parameters)
        Job.__init__(self, "POST", self.target, body)


class Batch(object):
    """ A collection of :class:`.Job` objects that can be submitted
    to a :class:`.BatchResource`. References to previous jobs are only
    valid **within the same batch** and will not work across batches.
    """

    #: The graph with which this batch is associated
    graph = None

    def __init__(self, graph):
        self.graph = graph
        self.jobs = []

    def __getitem__(self, index):
        return self.jobs[index]

    def __len__(self):
        return len(self.jobs)

    def __bool__(self):
        return bool(self.jobs)

    def __nonzero__(self):
        return bool(self.jobs)

    def __iter__(self):
        return iter(self.jobs)

    def resolve(self, node):
        """ Convert any references to previous jobs within the same batch
        into NodePointer objects.
        """
        if isinstance(node, Job):
            return NodePointer(self.find(node))
        else:
            return node

    def append(self, job):
        """ Add a job to this batch.

        :param job: A :class:`.Job` object to add to this batch.
        :rtype: :class:`.Job`

        """
        self.jobs.append(job)
        return job

    def find(self, job):
        """ Find the position of a job within this batch.
        """
        for i, candidate_job in pendulate(self.jobs):
            if candidate_job == job:
                return i
        raise ValueError("Job not found in batch")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import print_function

import os
import sys

from py2neo.ext.neobox.core import Warehouse
from py2neo.util import ustr


HELP = """\
Usage: {script} list
       {script} install «name» «edition» «version»
       {script} remove «name» [force]
       {script} rename «name» «new_name»
       {script} start «name»
       {script} stop «name»
       {script} open «name»
       {script} drop «name» [force]

Manager for multiple Neo4j server installations (boxes), each running on
its own unique port.

Commands:
  list     List all installed boxes
  install  Install Neo4j into a box
  remove   Remove a box
  rename   Rename a box
  start    Start a database server
  stop     Stop a database server
  open     Open the browser for a running server
  drop     Drop a database store

Environment:
  NEOBOX_HOME - base directory for installed boxes

Report bugs to nigel@py2neo.org
"""


def _help(script):
    print(HELP.format(script=os.path.basename(script)))


def _list(*args):
    warehouse = Warehouse()
    for box in warehouse.boxes():
        print(box.name)


def main():
    script, args = sys.argv[0], sys.argv[1:]
    try:
        if args:
            command, args = args[0], args[1:]
            if command == "help":
                _help(script)
            elif command == "list":
                _list(*args)
            else:
                warehouse = Warehouse()
                name, args = args[0], args[1:]
                box = warehouse.box(name)
                if command == "install":
                    edition, version = args
                    box.install(edition, version)
                    webserver_port = warehouse._ports[name]
                    webserver_https_port = webserver_port + 1
                    print("Created server instance %r configured on ports %s and %s" % (
                        name, webserver_port, webserver_https_port))
                elif command == "remove" and not args:
                    box.remove()
                elif command == "remove" and args[0] == "force":
                    box.remove(force=True)
                elif command == "rename":
                    new_name, = args
                    box.rename(new_name)
                elif command == "start":
                    ps = box.server.start()
                    print(ps.service_root.uri)
                elif command == "stop":
                    box.server.stop()
                elif command == "open":
                    if box.server:
                        box.server.graph.open_browser()
                    else:
                        raise RuntimeError("Server %r is not running" % name)
                elif command == "drop" and not args:
                    box.server.store.drop()
                elif command == "drop" and args[0] == "force":
                    box.server.store.drop(force=True)
                elif command == "load":
                    path, args = args[0], args[1:]
                    if not args:
                        box.server.store.load(path)
                    elif args[0] == "force":
                        box.server.store.load(path, force=True)
                    else:
                        raise ValueError("Bad command or arguments")
                elif command == "save":
                    path, args = args[0], args[1:]
                    if not args:
                        box.server.store.save(path)
                    elif args[0] == "force":
                        box.server.store.save(path, force=True)
                    else:
                        raise ValueError("Bad command or arguments")
                else:
                    raise ValueError("Bad command or arguments")
        else:
            _help(script)
    except Exception as error:
        sys.stderr.write(ustr(error))
        sys.stderr.write("\n")
        sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
.. note::
   Neobox has been written to work on Linux and may not operate
   correctly - or at all - on other platforms.

Neobox is a command line tool and API for managing multiple Neo4j server
installations (boxes), each running on its own unique port. The command
line tool is installed as part of the py2neo setup script and can be used
as follows::

    $ neobox install my-box community 2.1.5
    Created server instance 'my-box' configured on ports 47470 and 47471

    $ neobox list
    my-box

    $ neobox start my-box
    http://localhost:47476/

    $ neobox stop my-box

    $ neobox drop my-box

    $ neobox remove my-box

All command line features are also available as API calls.
"""


from py2neo.ext.neobox.core import *


__all__ = ["Warehouse", "Box"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from io import StringIO

from py2neo.core import Node, LabelSet, PropertySet, Relationship
from py2neo.cypher.lang import CypherParameter, CypherWriter
from py2neo.util import ustr, xstr


__all__ = ["CypherTask", "CreateNode", "CreateRelationship", "MergeNode"]


class CypherTask(object):
    """ The `CypherTask` class can either be used directly or as
    a base class for more specific statement implementations.
    """

    def __init__(self, statement="", parameters=None, **kwparameters):
        self.__statement = statement
        self.__parameters = dict(parameters or {}, **kwparameters)

    def __repr__(self):
        return "<CypherTask statement=%r parameters=%r>" % (self.statement, self.parameters)

    def __str__(self):
        return xstr(self.statement)

    def __unicode__(self):
        return ustr(self.statement)

    @property
    def statement(self):
        """ The Cypher statement.
        """
        return self.__statement

    @property
    def parameters(self):
        """ Dictionary of parameters.
        """
        return self.__parameters


class CreateNode(CypherTask):
    """ :class:`.CypherTask` for creating nodes.
    """

    def __init__(self, *labels, **properties):
        CypherTask.__init__(self)
        self.__node = Node(*labels, **properties)
        self.__return = False

    @property
    def labels(self):
        """ The full set of labels to apply to the created node.

        :rtype: :class:`py2neo.LabelSet`
        """
        return self.__node.labels

    @property
    def properties(self):
        """ The full set of properties to apply to the created node.

        :rtype: :class:`py2neo.PropertySet`
        """
        return self.__node.properties

    def set(self, *labels, **properties):
        """ Extra labels and properties to apply to the node.
        """
        self.__node.labels.update(labels)
        self.__node.properties.update(properties)
        return self

    def with_return(self):
        """ Include a RETURN clause in the statement.
        """
        self.__return = True
        return self

    @property
    def statement(self):
        """ The full Cypher statement.
        """
        string = StringIO()
        writer = CypherWriter(string)
        writer.write_literal("CREATE ")
        writer.write_node(self.__node, "a" if self.__return else None,
                          CypherParameter("P") if self.__node.properties else None)
        if self.__return:
            writer.write_literal(" RETURN a")
        return string.getvalue()

    @property
    def parameters(self):
        """ Dictionary of parameters.
        """
        if self.__node.properties:
            return {"P": self.properties}
        else:
            return {}


class CreateRelationship(CypherTask):
    """ :class:`.CypherTask` for creating nodes.
    """

    def __init__(self, *type_, **properties):
        CypherTask.__init__(self)
        self.__relationship = Relationship(*labels, **properties)
        self.__return = False

    @property
    def type(self):
        """ The full set of labels to apply to the created node.

        :rtype: :class:`py2neo.LabelSet`
        """
        return self.__relationship.type

    @property
    def properties(self):
        """ The full set of properties to apply to the created node.

        :rtype: :class:`py2neo.PropertySet`
        """
        return self.__relationship.properties

    def set(self, *labels, **properties):
        """ Extra labels and properties to apply to the node.
        """
        self.__relationship.labels.update(labels)
        self.__relationship.properties.update(properties)
        return self

    def with_return(self):
        """ Include a RETURN clause in the statement.
        """
        self.__return = True
        return self

    @property
    def statement(self):
        """ The full Cypher statement.
        """
        string = StringIO()
        writer = CypherWriter(string)
        writer.write_literal("CREATE ")
        writer.write_node(self.__relationship, "a" if self.__return else None,
                          CypherParameter("P") if self.__relationship.properties else None)
        if self.__return:
            writer.write_literal(" RETURN a")
        return string.getvalue()

    @property
    def parameters(self):
        """ Dictionary of parameters.
        """
        if self.__relationship.properties:
            return {"P": self.properties}
        else:
            return {}


class MergeNode(CypherTask):
    """ :class:`.CypherTask` for `merging <http://neo4j.com/docs/stable/query-merge.html>`_
    nodes.

    ::

        >>> from py2neo import Graph
        >>> graph = Graph()
        >>> tx = graph.cypher.begin()
        >>> tx.append(MergeNode("Person", "name", "Alice"))
        >>> tx.commit()
           | a
        ---+-----------------------
         1 | (n170 {name:"Alice"})


    """

    def __init__(self, primary_label, primary_key=None, primary_value=None):
        CypherTask.__init__(self)
        self.__node = Node(primary_label)
        if primary_key is not None:
            self.__node.properties[primary_key] = CypherParameter("V", primary_value)
        self.__labels = LabelSet()
        self.__properties = PropertySet()
        self.__return = False

    @property
    def labels(self):
        """ The full set of labels to apply to the merged node.

        :rtype: :class:`py2neo.LabelSet`
        """
        l = LabelSet(self.__labels)
        l.update(self.__node.labels)
        return l

    @property
    def properties(self):
        """ The full set of properties to apply to the merged node.

        :rtype: :class:`py2neo.PropertySet`
        """
        p = PropertySet(self.__properties)
        if self.primary_key:
            p[self.primary_key] = self.primary_value
        return p

    @property
    def primary_label(self):
        """ The label on which to merge.
        """
        return list(self.__node.labels)[0]

    @property
    def primary_key(self):
        """ The property key on which to merge.
        """
        try:
            return list(self.__node.properties.keys())[0]
        except IndexError:
            return None

    @property
    def primary_value(self):
        """ The property value on which to merge.
        """
        try:
            return list(self.__node.properties.values())[0].value
        except IndexError:
            return None

    def set(self, *labels, **properties):
        """ Extra labels and properties to apply to the node.

            >>> merge = MergeNode("Person", "name", "Bob").set("Employee", employee_id=1234)

        """
        self.__labels.update(labels)
        self.__properties.update(properties)
        return self

    def with_return(self):
        """ Include a RETURN clause in the statement.
        """
        self.__return = True
        return self

    @property
    def statement(self):
        """ The full Cypher statement.
        """
        string = StringIO()
        writer = CypherWriter(string)
        writer.write_literal("MERGE ")
        if self.__labels or self.__properties or self.__return:
            node_name = "a"
        else:
            node_name = None
        writer.write_node(self.__node, node_name)
        if self.__labels:
            writer.write_literal(" SET a")
            for label in self.__labels:
                writer.write_label(label)
        if self.__properties:
            writer.write_literal(" SET a={P}")
        if self.__return:
            writer.write_literal(" RETURN a")
        return string.getvalue()

    @property
    def parameters(self):
        """ Dictionary of parameters.
        """
        parameters = {}
        if self.__node.properties:
            parameters["V"] = self.primary_value
        if self.__properties:
            parameters["P"] = self.properties
        return parameters



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import unicode_literals

import json
import re
from xml.etree import ElementTree

from py2neo.cypher import cypher_repr
from py2neo.util import xstr


SIMPLE_NAME = re.compile(r"^[A-Za-z_][0-9A-Za-z_]*$")


def jsonify(obj, ensure_ascii=True):
    return json.dumps(obj, separators=(",", ":"), sort_keys=True, ensure_ascii=ensure_ascii)


def cyphify(obj, ensure_ascii=True):
    return cypher_repr(obj)


def _convert_xml(src, prefixes=None, verb=None, method=jsonify,
                 ensure_ascii=True, separator="\n"):
    """ Convert data from XML source to Cypher or Geoff.
    """
    TAG_PATTERN = re.compile(r"^(\{(.*)\})?(.*)$")
    if prefixes:
        prefixes = dict((b, a) for a, b in prefixes.items())
    nodes, rels, buffer, node_ids = [], [], [], []

    def local(tag):
        groups = TAG_PATTERN.match(tag).groups()
        if prefixes and groups[1] and groups[1] in prefixes:
            return prefixes[groups[1]] + "_" + groups[2]
        else:
            return groups[2]

    def node_no(node):
        if node not in nodes:
            n = len(nodes) + 1
            nodes.append(node)
            node_id = node.attrib.get("id")
            if node_id and (not ensure_ascii or SIMPLE_NAME.match(node_id)):
                node_ids.append(node_id)
            else:
                node_ids.append("node_{0}".format(n))
        return nodes.index(node)

    def walk(parent, child):
        if parent is not None:
            rels.append((node_no(parent), local(child.tag), node_no(child),
                         dict((key, value)
                              for key, value in child.attrib.items()
                              if key != "id")
            ))
        for grandchild in child:
            if len(grandchild) > 0:
                walk(child, grandchild)

    def typed(value):
        try:
            return int(value)
        except ValueError:
            try:
                return float(value)
            except ValueError:
                return value

    walk(None, ElementTree.fromstring(xstr(src)))
    for i, node in enumerate(nodes):
        properties = {}
        for child in node:
            is_leaf = len(child) == 0
            if child.text:
                inner_value = typed(child.text.strip())
            else:
                inner_value = None
            if is_leaf:
                if inner_value is not None:
                    properties[local(child.tag)] = inner_value
                for key, value in child.attrib.items():
                    if key != "id":
                        properties[local(child.tag) + " " + local(key)] = typed(value)
        if properties:
            buffer.append("({0} {1})".format(node_ids[i], method(properties, ensure_ascii=ensure_ascii)))
        else:
            buffer.append("({0})".format(node_ids[i]))
    for rel in rels:
        properties = rel[3]
        if properties:
            buffer.append("({0})-[:{1} {3}]->({2})".format(
                node_ids[rel[0]],
                rel[1] if SIMPLE_NAME.match(rel[1]) else method(rel[1], ensure_ascii=ensure_ascii),
                node_ids[rel[2]],
                method(properties, ensure_ascii=ensure_ascii),
            ))
        else:
            buffer.append("({0})-[:{1}]->({2})".format(
                node_ids[rel[0]],
                rel[1] if SIMPLE_NAME.match(rel[1]) else method(rel[1], ensure_ascii=ensure_ascii),
                node_ids[rel[2]],
            ))
    if verb:
        return verb + "\n" + separator.join(buffer)
    else:
        return separator.join(buffer)


def xml_to_geoff(src, prefixes=None):
    return _convert_xml(src, prefixes=prefixes, ensure_ascii=False)


def xml_to_cypher(src, prefixes=None):
    return _convert_xml(src, prefixes=prefixes, verb="CREATE", method=cyphify,
                        ensure_ascii=True, separator=",\n")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import print_function

import os
import sys

from py2neo import GraphError, Resource, Service, ServiceRoot
from py2neo.env import NEO4J_URI
from py2neo.packages.httpstream.numbers import UNPROCESSABLE_ENTITY
from py2neo.util import ustr


HELP = """\
Usage: {script} «user_name» «password» [«new_password»]

Authenticate against a Neo4j database server, optionally changing
the password.

Report bugs to nigel@py2neo.org
"""


class AuthenticationError(GraphError):
    """ Authentication failed with the supplied credentials.
    """


class PasswordChangeRequired(GraphError):
    """ A new password is required before authentication can occur.
    """


class Authentication(Service):
    """ Authentication management service.
    """

    @classmethod
    def for_service(cls, service_root):
        """ Fetch an Authentication instance for a given service root.

        :param service_root: A valid :class:`py2neo.ServiceRoot` instance.
        :rtype: :class:`.Authentication`
        """
        try:
            return cls(service_root.resource.metadata["authentication"])
        except KeyError:
            raise NotImplementedError("Authentication is not required for the service "
                                      "at %r" % service_root.uri.string)

    __instances = {}

    #: Instance metadata, updated on refresh.
    metadata = None

    def __new__(cls, uri):
        try:
            inst = cls.__instances[uri]
        except KeyError:
            inst = super(Authentication, cls).__new__(cls)
            inst.bind(uri)
            cls.__instances[uri] = inst
        return inst

    def refresh(self, user_name, password):
        """ Perform authentication to refresh the stored metadata.

        :arg user_name: Name of user to authenticate as.
        :arg password: The current password for this user.
        """
        try:
            self.metadata = self.resource.post({"username": user_name,
                                                "password": password}).content
        except GraphError as error:
            if error.response.status_code == UNPROCESSABLE_ENTITY:
                raise AuthenticationError("Cannot authenticate for user %r" % user_name)
            else:
                raise

    def authenticate(self, user_name, password, new_password=None):
        """ Authenticate to retrieve an auth token.

        :arg service_root: The :class:`py2neo.ServiceRoot` object requiring authentication.
        :arg user_name: Name of user to authenticate as.
        :arg password: The current password for this user.
        :arg new_password: A new password (optional, must not match the current password).
        :return: A valid auth token.
        :raise ValueError: If the new password supplied is invalid.
        :raise AuthenticationError: If the user cannot be authenticated.
        """
        self.refresh(user_name, password)
        if new_password is None:
            if self.metadata["password_change_required"]:
                raise PasswordChangeRequired("A password change is required for the service "
                                             "at %r" % self.service_root.uri.string)
            else:
                return self.metadata.get("authorization_token")
        else:
            password_manager = PasswordManager(self.metadata["password_change"])
            return password_manager.change(password, new_password)


class PasswordManager(Service):
    """ Password management service.
    """

    __instances = {}

    def __new__(cls, uri):
        try:
            inst = cls.__instances[uri]
        except KeyError:
            inst = super(PasswordManager, cls).__new__(cls)
            inst.bind(uri)
            cls.__instances[uri] = inst
        return inst

    def change(self, password, new_password):
        """ Change the authentication password.

        :arg password: The current password.
        :arg new_password: A new password (must not match the current password).
        :return: A valid auth token.
        :raise ValueError: If the new password is invalid.
        """
        try:
            response = self.resource.post({"password": password, "new_password": new_password})
        except GraphError as error:
            if error.response.status_code == UNPROCESSABLE_ENTITY:
                raise ValueError("Cannot change password")
            else:
                raise
        else:
            return response.content.get("authorization_token")


def get_auth_token(uri, user_name, password, new_password=None):
    """ Authenticate to retrieve an auth token.

    :arg uri: The root URI for the service requiring authentication.
    :arg user_name: Name of user to authenticate as.
    :arg password: The current password for this user.
    :arg new_password: A new password (optional, must not match the current password).
    :return: A valid auth token.
    :raise AuthenticationError: If the user cannot be authenticated.
    :raise ValueError: If the new password value is invalid.

    """
    auth = Authentication.for_service(ServiceRoot(uri))
    return auth.authenticate(user_name, password, new_password)


def _help(script):
    print(HELP.format(script=os.path.basename(script)))


def main():
    script, args = sys.argv[0], sys.argv[1:]
    try:
        if args:
            if 2 <= len(args) <= 3:
                print(get_auth_token(NEO4J_URI, *args))
            else:
                _help(script)
        else:
            _help(script)
    except Exception as error:
        sys.stderr.write(ustr(error))
        sys.stderr.write("\n")
        sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import codecs
import os
import sys

from py2neo.core import ServiceRoot
from py2neo.env import NEO4J_URI

from py2neo.ext.geoff import GeoffLoader


HELP = """\
Usage: {script} load «geoff-data»
       {script} load -f «geoff-file»
       {script} load -x «xml-data»
       {script} load -x -f «xml-file»

Load data into Neo4j from Geoff (or XML) data, optionally read from a file.

Environment:
  NEO4J_URI - base URI of Neo4j database, e.g. http://localhost:7474

Report bugs to nigel@py2neo.org
"""


def _help(script):
    sys.stderr.write(HELP.format(script=os.path.basename(script)))
    sys.stderr.write("\n")


def main():
    try:
        script, format_, args = sys.argv[0], sys.argv[1], sys.argv[2:]
    except IndexError:
        _help(sys.argv[0])
        sys.exit(1)

    uri = NEO4J_URI.resolve("/")
    service_root = ServiceRoot(uri.string)

    loader = GeoffLoader(service_root.graph)
    load = loader.load
    results = []

    try:
        while args:
            arg = args.pop(0)
            if arg.startswith("-"):
                if arg in ("-f", "--file"):
                    filename = args.pop(0)
                    with codecs.open(filename, encoding="utf-8") as f:
                        data = f.read()
                    results = load(data)
                elif arg in ("-x", "--xml"):
                    load = loader.load_xml
                else:
                    raise ValueError("Unknown option %s" % arg)
            else:
                results = load(arg)
            try:
                for result in results:
                    max_key_len = max(map(len, result.keys()))
                    for key in sorted(result):
                        reference = result.get_ref(key)
                        print("%s %s" % (key.ljust(max_key_len), reference))
            except TypeError:
                pass
    except Exception as error:
        sys.stderr.write(error.args[0])
        sys.stderr.write("\n")
        sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from py2neo.core import Graph, Node, NodePointer, Path, Relationship, Rev
from py2neo.cypher.lang import cypher_escape
from py2neo.cypher.util import StartOrMatch
from py2neo.util import ustr, xstr


__all__ = ["CreateStatement"]


def _(*args):
    return "".join("_" + ustr(arg) for arg in args)


class CreateStatement(object):
    """ Builder for a Cypher CREATE/CREATE UNIQUE statement.
    """

    #: The graph against which this statement is to be executed.
    graph = None

    #: The parameters to inject into this statement.
    parameters = None

    def __init__(self, graph):
        self.graph = graph
        self.supports_node_labels = self.graph.supports_node_labels
        self.entities = []
        self.names = []
        self.start_or_match_clause = StartOrMatch(self.graph)
        self.create_clause = []
        self.create_unique_clause = []
        self.return_clause = []
        self.parameters = {}

    def __repr__(self):
        return self.string

    def __str__(self):
        return xstr(self.__unicode__())

    def __unicode__(self):
        return self.string

    def __contains__(self, entity):
        return any(e is entity for e in self.entities)

    @property
    def string(self):
        """ The full Cypher statement as a string.
        """
        clauses = [self.start_or_match_clause.string]
        if self.create_clause:
            clauses.append("CREATE " + ",".join(self.create_clause))
        if self.create_unique_clause:
            clauses.append("CREATE UNIQUE " + ",".join(self.create_unique_clause))
        if self.return_clause:
            clauses.append("RETURN " + ",".join(self.return_clause))
        return "\n".join(clauses).strip()

    def post(self):
        return self.graph.cypher.post(self.string, self.parameters)

    def execute(self):
        """ Execute this statement.

        :return: A tuple of created entities.

        """
        if not self.entities:
            return ()
        raw = self.post().content
        columns = raw["columns"]
        data = raw["data"]
        dehydrated = dict(zip(columns, data[0]))
        for i, entity in enumerate(self.entities):
            node_names, rel_names = self.names[i]
            if isinstance(entity, Node):
                metadata = dehydrated[node_names[0]]
                entity.bind(metadata["self"], metadata)
            elif isinstance(entity, Relationship):
                metadata = dehydrated[rel_names[0]]
                entity.bind(metadata["self"], metadata)
            elif isinstance(entity, Path):
                for j, node in enumerate(entity.nodes):
                    metadata = dehydrated[node_names[j]]
                    node.bind(metadata["self"], metadata)
                for j, rel in enumerate(entity.rels):
                    metadata = dehydrated[rel_names[j]]
                    rel.bind(metadata["self"], metadata)
        return tuple(self.entities)

    def create(self, entity):
        """ Append an entity to the CREATE clause of this statement.

        :arg entity: The entity to create.

        """
        entity = Graph.cast(entity)
        index = len(self.entities)
        name = _(index)
        if isinstance(entity, Node):
            self.names.append(self._create_node(entity, name))
        elif isinstance(entity, Path):
            self.names.append(self._create_path(entity, name, unique=False))
        else:
            raise TypeError("Cannot create entity of type %s" % type(entity).__name__)
        self.entities.append(entity)

    def create_unique(self, entity):
        """ Append an entity to the CREATE UNIQUE clause of this statement.

        :arg entity: The entity to add.

        """
        entity = Graph.cast(entity)
        index = len(self.entities)
        name = _(index)
        if isinstance(entity, Path):
            if len(entity) == 0:
                raise ValueError("Cannot create unique path with zero length")
            if not any(node.bound or node in self for node in entity.nodes):
                raise ValueError("At least one node must be bound to create a unique path")
            self.names.append(self._create_path(entity, name, unique=True))
        else:
            raise TypeError("Cannot create unique entity of type %s" % type(entity).__name__)
        self.entities.append(entity)

    def _node_pattern(self, node, name, full):
        template = "{name}"
        kwargs = {"name": name}
        if full:
            if node.labels and self.supports_node_labels:
                template += "{labels}"
                kwargs["labels"] = "".join(":" + cypher_escape(label) for label in node.labels)
            if node.properties:
                template += " {{{name}}}"
                self.parameters[name] = node.properties
        return "(" + template.format(**kwargs) + ")"

    def _create_node(self, node, name):
        if node.bound:
            self.start_or_match_clause.node(name, "{" + name + "}")
            self.parameters[name] = node._id
        else:
            self.create_clause.append(self._node_pattern(node, name, full=True))
        self.return_clause.append(name)
        return [name], []

    def _create_path_nodes(self, path, name, unique):
        node_names = []
        for i, node in enumerate(path.nodes):
            if isinstance(node, NodePointer):
                node_names.append(_(node.address))
                # Switch out node with object from elsewhere in entity list
                nodes = list(path.nodes)
                try:
                    target_node = self.entities[node.address]
                except IndexError:
                    raise IndexError("Node pointer {%s} out of range" % node.address)
                if not isinstance(target_node, Node):
                    raise ValueError("Pointer {%s} does not refer to a node" % node.address)
                nodes[i] = target_node
                path._Path__nodes = tuple(nodes)
            elif node in self:
                node_name = _(self.entities.index(node))
                node_names.append(node_name)
            elif unique and not node.bound:
                node_name = name + "n" + ustr(i)
                node_names.append(node_name)
                self.return_clause.append(node_name)
            else:
                node_name = name + "n" + ustr(i)
                node_names.append(node_name)
                self._create_node(node, node_name)
        return node_names

    def _create_path(self, path, name, unique):
        node_names = self._create_path_nodes(path, name, unique)
        rel_names = []
        for i, rel in enumerate(path.rels):
            rel_name = name + "r" + ustr(i)
            rel_names.append(rel_name)
            if rel.bound:
                self.start_or_match_clause.relationship(rel_name, "{" + rel_name + "}")
                self.parameters[rel_name] = rel._id
            else:
                if rel.properties:
                    template = "{start}-[{name}:{type} {{{name}}}]->{end}"
                    self.parameters[rel_name] = rel.properties
                else:
                    template = "{start}-[{name}:{type}]->{end}"
                start_index, end_index = i, i + 1
                if isinstance(rel,  Rev):
                    start_index, end_index = end_index, start_index
                start_node = path.nodes[start_index]
                end_node = path.nodes[end_index]
                start = self._node_pattern(start_node, node_names[start_index],
                                           full=(unique and not start_node.bound and start_node not in self))
                end = self._node_pattern(end_node, node_names[end_index],
                                         full=(unique and not end_node.bound and end_node not in self))
                kwargs = {"start": start, "name": rel_name,
                          "type": cypher_escape(rel.type), "end": end}
                if unique:
                    self.create_unique_clause.append(template.format(**kwargs))
                else:
                    self.create_clause.append(template.format(**kwargs))
            self.return_clause.append(rel_name)
        return node_names, rel_names



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" The ogm module provides Object to Graph Mapping features similar to ORM
facilities available for relational databases. All functionality is available
through the :class:`Store` class which is bound to a specific
:class:`~py2neo.Graph` instance on creation.

Conceptually, a mapped object *owns* a single node within the graph along with
all of that node's outgoing relationships. These features are managed via a
pair of attributes called ``__node__`` and ``__rel__`` which store details of the
mapped node and the outgoing relationships respectively. The only specific
requirement for a mapped object is that it has a nullary constructor which can
be used to create new instances.

The ``__node__`` attribute holds a :class:`~py2neo.Node` object which is the
node to which this object is mapped. If the attribute does not exist, or is
:const:`None`, the object is considered *unsaved*.

The ``__rel__`` attribute holds a dictionary of outgoing relationship details.
Each key corresponds to a relationship type and each value to a list of
2-tuples representing the outgoing relationships of that type. Within each
2-tuple, the first value holds a dictionary of relationship properties (which
may be empty) and the second value holds the endpoint. The endpoint may be
either a :class:`~py2neo.Node` instance or another mapped object. Any such
objects which are unsaved will be lazily saved as required by creation of the
relationship itself. The following data structure outline shows an example of
a ``__rel__`` attribute (where `alice` and `bob` represent other mapped objects::

    {
        "LIKES": [
            ({}, alice),
            ({"since": 1999}, bob)
        ]
    }

To manage relationships, use the :func:`Store.relate` and :func:`Store.separate`
methods. Neither method makes any calls to the database and operates only on the
local ``__rel__`` attribute. Changes must be explicitly saved via one of the
available save methods. The :func:`Store.load_related` method loads all objects
marked as related by the ``__rel__`` attribute.

The code below shows an example of usage::

    from py2neo import Graph
    from py2neo.ext.ogm import Store

    class Person(object):

        def __init__(self, email=None, name=None, age=None):
            self.email = email
            self.name = name
            self.age = age

        def __str__(self):
            return self.name

    graph = Graph()
    store = Store(graph)

    alice = Person("alice@example.com", "Alice", 34)
    store.save_unique("People", "email", alice.email, alice)

    bob = Person("bob@example.org", "Bob", 66)
    carol = Person("carol@example.net", "Carol", 42)
    store.relate(alice, "LIKES", bob)     # these relationships are not saved
    store.relate(alice, "LIKES", carol)   # until `alice` is saved
    store.save(alice)

    friends = store.load_related(alice, "LIKES", Person)
    print("Alice likes {0}".format(" and ".join(str(f) for f in friends)))

"""


from py2neo.ext.ogm.store import *


__all__ = ["Store", "NotSaved"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from py2neo.legacy.index import Index
from py2neo.core import Node, Relationship, Resource, PropertyContainer, Service
from py2neo.packages.jsonstream import assembled


__all__ = ["LegacyResource", "LegacyNode"]


class LegacyResource(Service):

    __instances = {}

    def __new__(cls, uri):
        try:
            inst = cls.__instances[uri]
        except KeyError:
            inst = super(LegacyResource, cls).__new__(cls)
            inst.bind(uri)
            inst._indexes = {Node: {}, Relationship: {}}
            inst.supports_index_uniqueness_modes = inst.graph.neo4j_version >= (1, 9)
            cls.__instances[uri] = inst
        return inst

    def _index_manager(self, content_type):
        """ Fetch the index management resource for the given `content_type`.

        :param content_type:
        :return:
        """
        if content_type is Node:
            uri = self.resource.metadata["node_index"]
        elif content_type is Relationship:
            uri = self.resource.metadata["relationship_index"]
        else:
            raise TypeError("Indexes can manage either Nodes or Relationships")
        return Resource(uri)

    def get_indexes(self, content_type):
        """ Fetch a dictionary of all available indexes of a given type.

        :param content_type: either :py:class:`neo4j.Node` or
            :py:class:`neo4j.Relationship`
        :return: a list of :py:class:`Index` instances of the specified type
        """
        index_manager = self._index_manager(content_type)
        index_index = index_manager.get().content
        if index_index:
            self._indexes[content_type] = dict(
                (key, Index(content_type, value["template"]))
                for key, value in index_index.items()
            )
        else:
            self._indexes[content_type] = {}
        return self._indexes[content_type]

    def get_index(self, content_type, index_name):
        """ Fetch a specific index from the current database, returning an
        :py:class:`Index` instance. If an index with the supplied `name` and
        content `type` does not exist, :py:const:`None` is returned.

        :param content_type: either :py:class:`neo4j.Node` or
            :py:class:`neo4j.Relationship`
        :param index_name: the name of the required index
        :return: an :py:class:`Index` instance or :py:const:`None`

        .. seealso:: :py:func:`get_or_create_index`
        .. seealso:: :py:class:`Index`
        """
        if index_name not in self._indexes[content_type]:
            self.get_indexes(content_type)
        if index_name in self._indexes[content_type]:
            return self._indexes[content_type][index_name]
        else:
            return None

    def get_or_create_index(self, content_type, index_name, config=None):
        """ Fetch a specific index from the current database, returning an
        :py:class:`Index` instance. If an index with the supplied `name` and
        content `type` does not exist, one is created with either the
        default configuration or that supplied in `config`::

            # get or create a node index called "People"
            people = graph.get_or_create_index(neo4j.Node, "People")

            # get or create a relationship index called "Friends"
            friends = graph.get_or_create_index(neo4j.Relationship, "Friends")

        :param content_type: either :py:class:`neo4j.Node` or
            :py:class:`neo4j.Relationship`
        :param index_name: the name of the required index
        :return: an :py:class:`Index` instance

        .. seealso:: :py:func:`get_index`
        .. seealso:: :py:class:`Index`
        """
        index = self.get_index(content_type, index_name)
        if index:
            return index
        index_manager = self._index_manager(content_type)
        rs = index_manager.post({"name": index_name, "config": config or {}})
        index = Index(content_type, assembled(rs)["template"])
        self._indexes[content_type].update({index_name: index})
        return index

    def delete_index(self, content_type, index_name):
        """ Delete the entire index identified by the type and name supplied.

        :param content_type: either :py:class:`neo4j.Node` or
            :py:class:`neo4j.Relationship`
        :param index_name: the name of the index to delete
        :raise LookupError: if the specified index does not exist
        """
        if index_name not in self._indexes[content_type]:
            self.get_indexes(content_type)
        if index_name in self._indexes[content_type]:
            index = self._indexes[content_type][index_name]
            index.resource.delete()
            del self._indexes[content_type][index_name]
        else:
            raise LookupError("Index not found")

    def get_indexed_node(self, index_name, key, value):
        """ Fetch the first node indexed with the specified details, returning
        :py:const:`None` if none found.

        :param index_name: the name of the required index
        :param key: the index key
        :param value: the index value
        :return: a :py:class:`Node` instance
        """
        index = self.get_index(Node, index_name)
        if index:
            nodes = index.get(key, value)
            if nodes:
                return nodes[0]
        return None

    def get_or_create_indexed_node(self, index_name, key, value, properties=None):
        """ Fetch the first node indexed with the specified details, creating
        and returning a new indexed node if none found.

        :param index_name: the name of the required index
        :param key: the index key
        :param value: the index value
        :param properties: properties for the new node, if one is created
            (optional)
        :return: a :py:class:`Node` instance
        """
        index = self.get_or_create_index(Node, index_name)
        return index.get_or_create(key, value, properties or {})

    def get_indexed_relationship(self, index_name, key, value):
        """ Fetch the first relationship indexed with the specified details,
        returning :py:const:`None` if none found.

        :param index_name: the name of the required index
        :param key: the index key
        :param value: the index value
        :return: a :py:class:`Relationship` instance
        """
        index = self.get_index(Relationship, index_name)
        if index:
            relationships = index.get(key, value)
            if relationships:
                return relationships[0]
        return None


class LegacyNode(Node):
    """ Legacy Node object for pre-2.0 servers. Does not
    synchronise label information with server.
    """

    @property
    def labels(self):
        return self._Node__labels

    def bind(self, uri, metadata=None):
        PropertyContainer.bind(self, uri, metadata)

    def unbind(self):
        PropertyContainer.unbind(self)

    def pull(self):
        self._PropertyContainer__properties.clear()
        self.refresh()

    def push(self):
        PropertyContainer.push(self)

    def refresh(self):
        from py2neo.cypher.util import StartOrMatch
        graph = self.graph
        statement = StartOrMatch(graph).node("a", "{a}").string + "RETURN a"
        graph.cypher.execute(statement, {"a": self})
        self._Node__stale.clear()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from py2neo import ServerPlugin


__all__ = ["Gremlin"]


class Gremlin(ServerPlugin):
    """ Extension for Gremlin script execution. This requires the
    Gremlin server plugin.
    """

    def __init__(self, graph):
        super(Gremlin, self).__init__(graph, "GremlinPlugin")

    def execute(self, script):
        """ Execute a Gremlin script.
        """
        response = self.resources["execute_script"].post({"script": script})
        return self.graph.hydrate(response.content)
